from abc import ABC, abstractmethod

from .table import Table


class Database(ABC):
    """
    This is a base class that provides the basic functionality
    necessary for an AutoDB implementation.
    """

    @abstractmethod
    def __len__(self):
        """
        Should return the sum of all the objects
        stored in this database.
        :return:
        """
        raise NotImplemented

    @abstractmethod
    def __iter__(self):
        """
        Should allow iteration of the tables contained
        within this database.
        :return:
        """
        raise NotImplemented

    @abstractmethod
    def __repr__(self):
        """
        Should print out a summary of all of the object
        tables.
        :return:
        """
        raise NotImplemented

    @property
    @abstractmethod
    def tables(self):
        """
        Should return a view of all of the tables in this database.
        :return:
        """
        raise NotImplemented

    @abstractmethod
    def insert(self, item):
        """
        Inserts and indexes arbitrary Python objects.
        :param item:
        :return:
        """
        raise NotImplemented

    @abstractmethod
    def select(self, cls) -> Table:
        """
        Select items of type `cls`.
        :param cls:
        :return:
        """
        raise NotImplemented



import os
import shutil
import stat


def _remove_readonly(func, path, _):
    """Clear the readonly bit and reattempt the removal."""
    os.chmod(path, stat.S_IWRITE)
    func(path)


def rmdir(directory):
    """Removes the given directory entirely."""
    shutil.rmtree(directory, onerror=_remove_readonly)


def empty_directory(directory) -> None:
    """Removes the contents of a directory."""
    with os.scandir(directory) as dir_contents:
        for entry in dir_contents:
            if entry.is_file():
                os.unlink(entry)
            if entry.is_dir():
                rmdir(entry)



from abc import ABC, abstractmethod
from typing import Optional, Set


class IndexManager(ABC):

    @abstractmethod
    def index_item(self, item: object, index: int) -> None:
        """
        Inserts/creates index tables based on the given item and index.
        :param item
        :param index
        """
        raise NotImplemented

    @abstractmethod
    def unindex_item(self, item: object, index: int) -> None:
        """
        Removes indexes based on the given item and index.
        :param item:
        :param index:
        :return:
        """
        raise NotImplemented

    @abstractmethod
    def retrieve(self, **kwargs) -> Optional[Set[int]]:
        """
        Retrieves item indexes based on the given search parameters.
        :param kwargs:
        :return:
        """
        raise NotImplemented



from os import scandir, listdir, DirEntry, path
from typing import Dict

from ..errors import DatabaseNotFound


def load_tables(directory: str) -> str:
    """Scans the given directory for tables and returns any valid table directories."""
    tables = 0
    with scandir(path=directory) as curdir:
        for entry in curdir:
            if is_table(entry):
                tables += 1
                yield entry
    if tables == 0:
        raise DatabaseNotFound


def is_table(directory: DirEntry) -> bool:
    """Checks if a directory contains a table."""
    return directory.is_dir(follow_symlinks=False) and valid_table_contents(directory.path)


def is_shard_file(file: DirEntry) -> bool:
    """Checks if a file is a shard."""
    return is_shard(file.name) and file.is_file(follow_symlinks=False)


def is_index_file(file: DirEntry) -> bool:
    """Checks if a file is an index."""
    return is_index(file.name) and file.is_file(follow_symlinks=False)


def is_info_file(file: DirEntry) -> bool:
    """Checks if a file is an info file."""
    return is_info(file.name) and file.is_file(follow_symlinks=False)


def is_shard(file_name: str) -> bool:
    """Determines if a filename is a proper shard name."""
    return file_name.startswith("shard") and len(file_name) > 5 and file_name[5:].isdigit()


def is_index(file_name: str) -> bool:
    """Determines if a filename is a proper index name."""
    return file_name == "index"


def is_info(file_name: str) -> bool:
    """Determines if a filename is a proper info name."""
    return file_name == "info"


def create_index_path(directory: str) -> str:
    """Creates an index path for a given directory."""
    return path.join(directory, "index")


def create_info_path(directory: str) -> str:
    """Creates an info path for a given directory."""
    return path.join(directory, "info")


def create_shard_path(directory: str, shard_number: int) -> str:
    """Creates a shard path for a given directory and shard number."""
    return f"""{path.join(directory, "shard")}{shard_number}"""


def get_shard_number(file_name: str) -> int:
    """Retrieves the shard number from a shard filename."""
    return int(file_name[5:])


def get_shard_file_paths(directory: str) -> Dict[int, str]:
    """Gets all of the shard files in a table directory."""
    return dict((get_shard_number(file), path.join(directory, file)) for file in
                [file for file in listdir(directory) if is_shard(file)])


def valid_table_contents(dir_path: str) -> bool:
    """Validates that a table has all of the necessary information to load data."""
    files = [file for file in listdir(dir_path)]
    index_files = [file for file in files if is_index(file)]
    shard_files = [file for file in files if is_shard(file)]
    info_files = [file for file in files if is_info(file)]
    if len(index_files) < 1 or len(shard_files) < 1 or len(info_files) < 1:
        return False
    return True



from typing import Dict


def retrieve_possible_object_indexes(complex_object: object) -> Dict[str, object]:
    """
    This method retrieves public variables of a class
    that can be used to build indexes.
    :param complex_object:
    :return:
    """

    indexes: Dict[str, object] = {}
    object_vars = vars(complex_object)
    for var in object_vars:
        if not var.startswith("_"):
            var_value = getattr(complex_object, var)
            indexes.update({var: var_value})
    return indexes



import pickle
from typing import List, Optional, Generator, Set

from litedb.abc.table import Table
from ..index.memory_index import MemoryIndex


class MemoryTable(Table):
    """
    This class stores elements of a certain class type within a regular Python list.
    Object attributes are indexed and stored within indexes that can then be used to retrieve
    objects from the list.
    """

    def __init__(self) -> None:
        self.size = 0
        self.table: List[object] = []
        self.pickle_table: List[Optional[bytes]] = []
        self.unused_indexes: Set[int] = set()
        self.index_manager = MemoryIndex()

    def __repr__(self):
        return f"Table(size={self.size})"

    def __len__(self):
        """Returns the number of items stored in this table."""
        return self.size

    def _insert(self, item: object) -> None:
        """Internal method that inserts an item into the pickle table."""
        byte_repr = pickle.dumps(item, protocol=pickle.HIGHEST_PROTOCOL)
        if len(self.unused_indexes) > 0:
            index = self.unused_indexes.pop()
            self.table[index] = item
            self.pickle_table[index] = byte_repr
        else:
            index = len(self.table)
            self.table.append(item)
            self.pickle_table.append(byte_repr)
        self.size += 1
        self.index_manager.index_item(item, index)

    def retrieve(self, **kwargs) -> [Generator[object, None, None]]:
        """Retrieves all items that match the given parameters."""
        if len(kwargs) == 0:
            raise ValueError
        indexes = self.index_manager.retrieve(**kwargs)
        if indexes:
            return (pickle.loads(self.pickle_table[index]) for index in indexes)
        else:
            # noinspection PyRedundantParentheses
            return ([])  # Disabling this inspection because we need to return a generator

    def retrieve_all(self) -> [Generator[object, None, None]]:
        """Retrieves all items in this table."""
        return (pickle.loads(item) for index, item in enumerate(self.pickle_table) if index not in self.unused_indexes)

    @property
    def indexes(self) -> List[str]:
        """Returns a list of all of the indexes in this table."""
        return list(self.index_manager.index_map.keys())

    def delete(self, **kwargs) -> None:
        """Delete items that match the given parameters."""
        if len(kwargs) == 0:
            raise ValueError
        indexes_to_delete = self.index_manager.retrieve(**kwargs)
        self._delete(indexes_to_delete)

    def _delete(self, indexes) -> None:
        """Internal method to remove the given indexes from the table and indexes."""
        if indexes:
            for index in indexes:
                item = self.table[index]
                self.index_manager.unindex_item(item, index)
                self.table[index] = None
                self.pickle_table[index] = None
            self.unused_indexes.update(indexes)
            self.size -= len(indexes)

    def clear(self) -> None:
        """Removes all items from this table."""
        indexes_to_delete = (index for index in range(len(self.table)) if index not in self.unused_indexes)
        self._delete(indexes_to_delete)



from typing import Optional, Set

from litedb.abc import IndexManager
from .index import Index
from ..errors import InvalidRange
from ..utils.index import retrieve_possible_object_indexes

NoneType = type(None)


class MemoryIndex(IndexManager):
    """This is a index manager that handles indexes for all of the different types
    in the database."""

    def __init__(self):
        self.index_map = {}
        self.index_blacklist = set()

    def index_item(self, item: object, index: int) -> None:
        """Inserts/creates index tables based on the given object."""
        indexes = retrieve_possible_object_indexes(item)
        for var_name, value in indexes.items():
            if var_name in self.index_blacklist:
                continue
            if var_name not in self.index_map:
                # if the first item value is None, create the index without assigning type
                value_type = type(value)
                if value_type is NoneType:
                    self.index_map.update({var_name: Index()})
                else:
                    self.index_map.update({var_name: Index(type(value))})
            try:
                self.index_map[var_name].add(value, index)
            except TypeError:
                self.index_map.pop(var_name)
                self.index_blacklist.add(var_name)

    def unindex_item(self, item: object, index: int) -> None:
        """Removes indexes for the given object."""
        indexes = retrieve_possible_object_indexes(item)
        for var_name, value in indexes.items():
            if var_name not in self.index_blacklist:
                self.index_map[var_name].destroy(value, index)

    def retrieve(self, **kwargs) -> Optional[Set[int]]:
        """Retrieves indexes that match the given parameters."""
        indexes: Set[int] = set()
        for x, key in enumerate(kwargs.keys()):
            if key in self.index_blacklist or key not in self.index_map:
                raise IndexError(f"{key} is not a valid index!")
            index = self.index_map[key]
            if len(index) == 0:
                continue
            value = kwargs[key]
            if isinstance(value, tuple):
                if len(value) != 2:
                    raise InvalidRange
                low, high = value
                if low is not None and not isinstance(low, index.index_type):
                    raise ValueError(f"The low value of \"{key}\" must be of type {index.index_type}")
                if high is not None and not isinstance(high, index.index_type):
                    raise ValueError(f"The high value of \"{key}\" must be of type {index.index_type}")
                if x == 0:
                    results = index.retrieve_range(low, high)
                    if results is not None:
                        indexes.update(results)
                else:
                    results = index.retrieve_range(low, high)
                    if results is not None:
                        indexes.intersection_update(results)
            else:
                if value is not None and not isinstance(value, index.index_type):
                    raise ValueError(f"\"{key}\" must be of type {index.index_type}")
                results = index.retrieve(value)
                if results is not None:
                    if x == 0:
                        indexes.update(results)
                    else:
                        indexes.intersection_update(results)
        if len(indexes) > 0:
            return indexes



import os
from typing import Dict, ValuesView

from litedb.abc.database import Database
from ..database.config import Config
from ..errors import DatabaseNotFound
from ..table import PersistentTable
from ..utils.path import load_tables


class DiskDatabase(Database):
    """Persistent implementation of the AutoDB interface."""

    def __init__(self, directory: str, config: Config = Config()):
        self._tables: Dict[object, PersistentTable] = {}
        self.directory = directory
        self._config = config
        if not os.path.exists(directory):
            os.mkdir(directory)
        if os.path.exists(directory):
            try:
                for table in load_tables(directory):
                    table = PersistentTable._from_file(table)
                    self._tables.update({table._table_type: table})
            except DatabaseNotFound:
                pass

    def __iter__(self):
        """Returns the various table types that this database contains."""
        return iter(self._tables.keys())

    def __len__(self):
        """Returns the number of objects contained in this database."""
        return sum(len(table) for table in self._tables.values())

    def __repr__(self):
        return f"DiskDatabase({self.directory})"

    @property
    def tables(self) -> ValuesView[PersistentTable]:
        """Returns a view of all of the tables present in this database."""
        return self._tables.values()

    def insert(self, item: object) -> None:
        """Inserts an arbitrary Python class into the database. Do not use this
        database to store raw types."""
        class_name = type(item)
        try:
            self._tables[class_name]._insert(item)
        except KeyError:
            self._tables.update(
                {class_name: PersistentTable._new(self._config, os.path.join(self.directory, hex(abs(hash(class_name)))),
                                                  table_type=class_name)})
            self._tables[class_name]._insert(item)

    def batch_insert(self, items: list) -> None:
        """Inserts a list of similar Python class into the database."""
        if len(items) == 0:
            return
        else:
            first_item_type = type(items[0])
            try:
                self._tables[first_item_type]._batch_insert(items)
            except KeyError:
                self._tables.update(
                    {first_item_type: PersistentTable._new(self._config, os.path.join(self.directory, first_item_type.__name__),
                                                           table_type=first_item_type)})
                self._tables[first_item_type]._batch_insert(items)

    def select(self, cls):
        """Retrieves the table that contains classes of the given type."""
        if cls in self._tables:
            return self._tables[cls]
        else:
            raise KeyError(f"No table of {cls} exists in this database!")

    def commit(self):
        """Commits all data in this database to disk."""
        for table in self._tables.values():
            table.commit()



import os
from typing import List, Generator, Union, Set

from sortedcontainers import SortedList, SortedDict

from litedb.abc.table import Table
from ..index import PersistentIndex
from ..shard import ShardManager
from ..database.config import Config
from ..utils.io import empty_directory
from ..utils.path import create_info_path, create_index_path
from ..utils.serialization import load_object, dump_object


class PersistentTable(Table):
    """
    This class stores elements of a certain class type within a regular Python list.
    Object attributes are indexed and stored within indexes that can then be used to retrieve
    objects from the list.
    """

    def __init__(self, config: Config = None, directory: str = None,
                 table_type=None) -> None:
        """
        This class can be instantiated as either a fresh new table or as an existing one from a file structure.
        To create a new table, an empty table directory and table type must be specified. Otherwise path info for
        the existing table must be supplied.
        :param directory:
        :param table_type:
        """

        self._directory = directory
        self._modified = False
        if not os.path.exists(directory):
            os.mkdir(directory)
        self._info_path = create_info_path(self._directory)

        if table_type is not None:
            self._table_type = table_type
            self._size = 0
            self._unused_indexes: SortedList = SortedList()
            self._config = config
        else:
            self._table_type = load_object(os.path.join(self._info_path, "table_type"))
            self._size = load_object(os.path.join(self._info_path, "size"))
            # noinspection PyTypeChecker
            self._unused_indexes: SortedList = load_object(os.path.join(self._info_path, "unused_indexes"))
            # noinspection PyTypeChecker
            self._config: Config = load_object(os.path.join(self._info_path, "config"))

        self._index_path = create_index_path(self._directory)
        self._shard_manager = ShardManager(self._directory, self._config)
        self._index_manager: PersistentIndex = PersistentIndex(self._index_path)

    def __repr__(self):
        return f"Table(size={self._size})"

    def __len__(self):
        return self._size

    @classmethod
    def _from_file(cls, directory: str):
        """Internal method to load a table from disk."""
        return cls(directory=directory)

    @classmethod
    def _new(cls, config: Config, directory: str, table_type):
        """Creates a new table with the given directory as the persistence location."""
        return cls(config=config, directory=directory, table_type=table_type)

    def commit(self) -> None:
        """Commits any changes made to this table to disk."""
        if self._modified:
            dump_object(os.path.join(self._info_path, "table_type"), self._table_type)
            dump_object(os.path.join(self._info_path, "size"), self._size)
            dump_object(os.path.join(self._info_path, "unused_indexes"), self._unused_indexes)
            dump_object(os.path.join(self._info_path, "config"), self._config)
            self._index_manager.commit()
            self._shard_manager.commit()
            self._modified = False

    def _insert(self, item: object) -> None:
        """Internal method that indexes and stores an object."""
        if len(self._unused_indexes) > 0:
            index = self._unused_indexes.pop()
        else:
            index = self._size
            item_dict = SortedDict()
            item_dict[index] = item
            self._shard_manager.insert(item_dict)
        self._size += 1
        self._index_manager.index_item(item, index)
        self._modified = True

    def _batch_insert(self, item_list: List[object]) -> None:
        """Internal method that indexes and stores a list of objects."""
        first_item_type = type(item_list[0])
        item_dict = SortedDict()
        for item in item_list:
            if type(item) != first_item_type:
                raise ValueError("Batch insert requires all elements to be of the same type!")
            if len(self._unused_indexes) > 0:
                index = self._unused_indexes.pop()
            else:
                index = self._size
            self._size += 1
            item_dict[index] = item
        self._shard_manager.insert(item_dict)
        for index, item in item_dict.items():
            self._index_manager.index_item(item, index)
        self._modified = True

    def retrieve(self, **kwargs) -> [Generator[object, None, None]]:
        """Retrieves items in this table based on the given argument descriptors."""
        if len(kwargs) == 0:
            raise ValueError
        indexes = self._index_manager.retrieve(**kwargs)
        if indexes:
            return self._shard_manager.retrieve(indexes)
        else:
            return ([])

    def retrieve_all(self) -> [Generator[object, None, None]]:
        """Retrieves all items in this table."""
        return self._shard_manager.retrieve_all()

    @property
    def indexes(self) -> List[str]:
        """Returns a list of all the indexes in this table."""
        return list(self._index_manager.index_map.keys())

    @property
    def modified(self) -> bool:
        """Indicates whether this table has unsaved changes."""
        return self._modified

    @property
    def size(self):
        """Return the number of items in this table."""
        return self._size

    def delete(self, **kwargs):
        """Removes items from this table based on the given descriptors."""
        if len(kwargs) == 0:
            raise ValueError
        indexes_to_delete = self._index_manager.retrieve(**kwargs)
        if indexes_to_delete:
            self._modified = True
            self._delete_indexes(indexes_to_delete)

    def _delete_indexes(self, indexes: Union[Set[int], List[int]]) -> None:
        """Internal method to actually perform deletion."""
        indexes_to_delete = sorted(indexes)  # Sort indexes for shards
        self._size -= len(indexes_to_delete)  # Decrement the size of the table
        if indexes_to_delete:
            self._modified = True
            items_to_delete = self._shard_manager.retrieve(indexes_to_delete)
            for item, index in zip(items_to_delete, indexes_to_delete):
                self._index_manager.unindex_item(item, index)
            self._unused_indexes.update(indexes_to_delete)
            self._shard_manager.delete(indexes_to_delete)

    def clear(self):
        """Removes any stored data relating to this table and clears out all items from this table."""
        empty_directory(self._directory)
        self._shard_manager = ShardManager(self._directory, self._config)
        self._index_manager = PersistentIndex(self._index_path)
        self._size = 0
        self._unused_indexes: SortedList = SortedList()



from abc import ABC, abstractmethod


class Table(ABC):

    @abstractmethod
    def __repr__(self):
        raise NotImplemented

    @abstractmethod
    def __len__(self):
        raise NotImplemented

    @abstractmethod
    def retrieve(self, **kwargs):
        """Filters search results."""
        raise NotImplemented

    @abstractmethod
    def retrieve_all(self):
        """Returns a generator over all of the items in the table."""
        raise NotImplemented

    @property
    @abstractmethod
    def indexes(self):
        """Retrieves all of the valid indexes for this type."""
        raise NotImplemented

    @abstractmethod
    def delete(self, **kwargs):
        """Deletes all items in this table based on the parameters."""
        raise NotImplemented

    @abstractmethod
    def clear(self):
        """Removes all items from this table."""
        raise NotImplemented



import os
import pickle
import sys
from io import BytesIO
from typing import Optional

from ..errors import PathError


def serialize(item: object) -> bytes:
    """Serializes the given object using pickle."""
    return pickle.dumps(item, pickle.HIGHEST_PROTOCOL)


def deserialize(raw_data: bytes):
    """Deserializes bytes to object using pickle."""
    return pickle.loads(raw_data)


def load_object(path: str) -> object:
    """Load python object from disk using pickle. Used to load indexes and other attributes."""
    if not os.path.exists(path) or not os.path.isfile(path):
        return
    with open(path, "rb") as file:
        return pickle.load(file)


def dump_object(path: str, item: object) -> None:
    """Persists object to disk using pickle. Used to save indexes and other attributes."""
    if not os.path.exists(os.path.dirname(path)):
        os.mkdir(os.path.dirname(path))
    with open(path, "wb") as file:
        pickle.dump(item, file, pickle.HIGHEST_PROTOCOL)


def load_shard(path: str) -> Optional[BytesIO]:
    """
    This function returns a serialized Python object.
    :param path:
    :return:
    """
    if not os.path.exists(path) or not os.path.isfile(path):
        return
    with open(path, "rb") as file:
        file.seek(0)
        bytes_io = BytesIO(file.read())
        bytes_io.seek(0)
        return bytes_io


def dump_shard(path: str, item: BytesIO) -> None:
    """
    This function saves an object with a checksum to disk.
    :param item:
    :param path:
    :return:
    """
    if not os.path.exists(os.path.dirname(path)):
        os.mkdir(os.path.dirname(path))
    with open(path, "wb") as file:
        file.seek(0)
        file.write(item.read())
        file.truncate()


def get_checksum(path: str) -> int:
    """
    This is a utility function that returns the checksum
    for a given serialized object.
    :param path:
    :return:
    """
    if not os.path.exists(path) or not os.path.isfile(path):
        raise PathError
    with open(path, "rb") as file:
        chksum_bytes = file.read(4)
    return int.from_bytes(chksum_bytes, sys.byteorder)



import pickle
from hashlib import blake2b
from zlib import adler32


def checksum(item: object) -> bytes:
    """Returns a checksum of the given object."""
    return blake2b(pickle.dumps(item, pickle.HIGHEST_PROTOCOL)).digest()


def chksum(byte_string: bytes) -> int:
    """Returns a fast checksum of the given byte string."""
    return adler32(byte_string)



from typing import Union, Optional, Set

from sortedcontainers import SortedDict

NoneType = type(None)


class Index:
    """
    This class stores maps valid index values to their corresponding list index.
    """

    def __init__(self, index_type=None):
        self.indexes: SortedDict[object, Union[int, Set[int]]] = SortedDict()
        self.none_indexes: Set[int] = set()
        self._index_type = index_type

    @property
    def index_type(self):
        return self._index_type

    def __eq__(self, other):
        if isinstance(other, Index):
            return self.none_indexes == other.none_indexes and self.indexes == other.indexes and self.index_type == other.index_type
        else:
            raise NotImplementedError

    def __len__(self):
        return len(self.indexes) + len(self.none_indexes)

    def add(self, value, index: int) -> None:
        """This method adds an index for a given value."""

        # If the index type is None, check the value for a type and assign the index type to it if it is not None
        if self._index_type is None:
            value_type = type(value)
            if value_type is not NoneType:
                self._index_type = value_type

        if value is None:
            self.none_indexes.add(index)
            return
        if value in self.indexes:
            indexes = self.indexes[value]
            if isinstance(indexes, set):
                self.indexes[value].add(index)
            else:
                self.indexes[value] = {indexes, index}
        else:
            self.indexes.update({value: index})

    def retrieve(self, value) -> Set[int]:
        """Return a set that contains the indexes that match the specified value."""
        if value is None:
            if len(self.none_indexes) > 0:
                return self.none_indexes
        elif value in self.indexes:
            indexes = self.indexes[value]
            if isinstance(indexes, int):
                return {indexes}
            return self.indexes[value]
        else:
            return set()

    def retrieve_range(self, low, high) -> Optional[Set[int]]:
        """This function retrieves a range of values depending on the high and low indexes given."""
        if low is None:
            return_set = self.none_indexes.copy()
            min_index = 0
        else:
            min_index = self.indexes.bisect_left(low)
            return_set = set()
        max_index = self.indexes.bisect_right(high) if high is not None else len(self.indexes)
        index_sets = self.indexes.values()[min_index:max_index]

        if len(index_sets) == 0 and len(return_set) == 0:
            return
        for index in index_sets:
            if isinstance(index, set):
                return_set.update(index)
            else:
                return_set.add(index)
        return return_set

    def destroy(self, value, index: int) -> None:
        """
        Removes an index from the database depending on the value of the tracked value.
        :param value:
        :param index:
        :return:
        """
        if value is None:
            self.none_indexes.remove(index)
            return
        entry = self.indexes[value] if value in self.indexes else None
        if entry is not None:
            if isinstance(entry, set):
                if index not in entry:
                    raise KeyError
                entry.remove(index)
                if len(entry) == 1:
                    self.indexes[value] = entry.pop()
            else:
                if index == entry:
                    self.indexes.pop(value)
                else:
                    raise KeyError
        else:
            raise KeyError



from collections import deque
from typing import Optional


class ShardLRU:
    """
    This is a most recently used cache that tracks when shards
    should be removed from memory.
    """

    def __init__(self, max_len=64) -> None:
        self.max_len: int = max_len
        self.mru = deque(maxlen=max_len + 1)

    def update(self, shard_index: int) -> Optional[int]:
        """Handles evicting old shards when a new one is added."""
        try:
            self.mru.remove(shard_index)
        except ValueError:
            pass
        self.mru.appendleft(shard_index)
        if len(self.mru) > self.max_len:
            return self.mru.pop()



from typing import Tuple, Generator, Iterable

from sortedcontainers import SortedDict

from .buffer import ShardBuffer
from ..utils.path import get_shard_file_paths
from ..database.config import Config


class ShardManager:
    """This class handles the high-level shard operations by manipulating the shard buffer."""

    def __init__(self, table_dir: str, config: Config) -> None:
        shard_paths = get_shard_file_paths(table_dir)
        self.buffer = ShardBuffer(table_dir, shard_paths, config)
        self.config = config

    def retrieve(self, indexes: Iterable[int]) -> Generator[object, None, None]:
        """
        Retrieves objects based on their indexes.
        :param indexes:
        :return:
        """
        shard_indexes = [self.calculate_shard_number(index) for index in indexes]
        shard_indexes.sort(key=lambda x: x[0])
        for shard, index in shard_indexes:
            yield self.buffer[shard][index]

    def retrieve_all(self) -> Generator[object, None, None]:
        """
        Retrieves all objects in the table.
        :return:
        """
        for shard in self.buffer:
            for byte in filter(lambda x: x is not None, shard):
                yield byte

    def insert(self, items: SortedDict) -> None:
        """Inserts and persists the given collection of items."""
        prepped_items = SortedDict()
        for key, value in items.items():
            prepped_items[self.calculate_shard_number(key)] = value
        for shard_index, value in prepped_items.items():
            shard, index = shard_index
            self.buffer[shard][index] = value

    def delete(self, indexes: Iterable[int]) -> None:
        """Removes the items with the given indexes."""
        shard_indexes = [self.calculate_shard_number(index) for index in indexes]
        shard_indexes.sort(key=lambda x: x[0])
        for shard, index in shard_indexes:
            self.buffer[shard][index] = None

    def commit(self):
        """Persists all data to disk."""
        self.buffer.commit()

    def calculate_shard_number(self, index: int) -> Tuple[int, int]:
        """Calculates the shard index and the item index within a shard."""
        return index // self.config.page_size, index % self.config.page_size



"""This class allows configuration of the page mechanism in liteDB."""


class Config:

    def __init__(self, page_size: int = 512, page_cache: int = 512):
        self._page_size = page_size
        self._page_cache = page_cache

    @property
    def page_size(self):
        return self._page_size

    @property
    def page_cache(self):
        return self._page_cache



import os

from .memory_index import MemoryIndex
from ..utils.serialization import dump_object, load_object


class PersistentIndex(MemoryIndex):
    """An extension of the in-memory index class that commits index
    changes to disk."""

    def __init__(self, index_path: str) -> None:
        super().__init__()
        self.index_path = index_path
        self.blacklist_path = os.path.join(self.index_path, "blacklist")
        self.map_path = os.path.join(self.index_path, "map")
        self.load()

    def load(self) -> None:
        """Loads the index from disk."""
        index_map = load_object(self.map_path)
        if index_map is not None:
            self.index_map = index_map
        blacklist = load_object(self.blacklist_path)
        if blacklist is not None:
            self.index_blacklist = blacklist

    def commit(self) -> None:
        """Persists the index to disk."""
        dump_object(self.blacklist_path, self.index_blacklist)
        dump_object(self.map_path, self.index_map)



import sys
from io import BytesIO
from typing import List, Optional

from ..utils.checksum import chksum
from ..utils.serialization import serialize, deserialize


class Shard:
    # Constants for None so they don't constantly have to be recalculated
    none_constant = b'\x80\x03N.'
    none_hash = chksum(none_constant)

    def __init__(self, shard_size: int = 512) -> None:
        self.max_size: int = shard_size
        self.checksum = 0
        self.binary_blobs: List[Optional[bytes]] = [None] * self.max_size

    def __getitem__(self, key: int) -> object:
        object_bytes = self.binary_blobs[key]
        return object_bytes if object_bytes is None else deserialize(object_bytes)

    def __setitem__(self, key: int, value: object) -> None:
        if value is None:  # We are removing this object from the database...
            # Null out the field in this shard and update the checksum
            self.binary_blobs[key] = self.none_constant
            self.checksum = self.checksum ^ self.none_hash
        else:
            # Convert the object to bytes, add it to the shard and update the checksum
            self.binary_blobs[key] = serialize(value)
            self.checksum = self.checksum ^ chksum(self.binary_blobs[key])

    @classmethod
    def from_bytes(cls, bytes: BytesIO, size: int = 512):
        """Converts the given BytesIO object into a Shard instance."""

        # Seek back to the beginning of this byte buffer
        bytes.seek(0)

        # Initialize the shard
        shard = cls(size)

        # Get the checksum
        num = int.from_bytes(bytes.read(4), byteorder=sys.byteorder, signed=False)
        shard.checksum = num

        # Alias inner arrays for faster lookups
        blobs = shard.binary_blobs

        for i in range(size):
            size_bytes = bytes.read(8)
            if size_bytes == b"":
                break
            object_size: int = int.from_bytes(size_bytes, byteorder=sys.byteorder, signed=False)
            object_bytes = bytes.read(object_size)
            blobs[i] = object_bytes

        return shard

    def to_bytes(self) -> BytesIO:
        """Converts this shard into a BytesIO instance that can then be written to disk."""

        # Create bytebuffer
        byte_buffer = BytesIO()

        # Convert checksum to bytes and write to buffer
        checksum = self.checksum.to_bytes(4, byteorder=sys.byteorder, signed=False)
        byte_buffer.write(checksum)

        # Convert stored objects to bytes and write to buffer
        for blob in self.binary_blobs:
            if blob is not None:
                object_size = len(blob)
                object_size = object_size.to_bytes(8, byteorder=sys.byteorder, signed=False)
                byte_buffer.write(object_size)
                byte_buffer.write(blob)

        # Seek back to the beginning of the byte buffer and return
        byte_buffer.seek(0)
        return byte_buffer



from typing import Dict, ValuesView

from litedb.abc.database import Database
from litedb.abc.table import Table
from ..table import MemoryTable


class MemoryDatabase(Database):
    """In memory implementation of the AutoDB interface."""

    def __init__(self) -> None:
        self._tables: Dict[object, MemoryTable] = {}
        self.selected_table = None

    def __len__(self):
        """Returns the number of objects contained in this database."""
        return sum((len(table) for table in self._tables.values()))

    def __repr__(self):
        return "".join([f"{str(cls)}: size={self._tables[cls].size}\n" for cls in self._tables.keys()])

    def __iter__(self):
        """Returns the various table types that this database contains."""
        return iter(self._tables.keys())

    @property
    def tables(self) -> ValuesView[Table]:
        """Returns a view of all of the tables in this database."""
        return self._tables.values()

    def insert(self, complex_object: object):
        """Inserts an arbitrary Python class into the database. Do not use this
        database to store raw types."""
        class_type = type(complex_object)
        if class_type not in self._tables:
            self._tables.update({class_type: MemoryTable()})
        self._tables[class_type]._insert(complex_object)

    def select(self, cls) -> Table:
        """Returns the table that contains classes of the given type."""
        if cls in self._tables:
            return self._tables[cls]
        else:
            raise KeyError(f"No table of {cls} exists in this database!")



import os

from setuptools import setup, find_packages


def read(file_name: str):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


setup(
    name="litedb",
    version="0.1.1",
    author="Jonathan Vusich",
    author_email="jonathanvusich@gmail.com",
    description="A pure Python NoSQL database",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="nosql database python",
    url="https://github.com/JonathanVusich/litedb",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=read("requirements.txt"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Database",
        "License :: OSI Approved :: MIT License",
    ],
)



import os
from typing import Dict

from .shard import Shard
from .shardlru import ShardLRU
from ..utils.serialization import load_shard, dump_shard, get_checksum
from ..database.config import Config


class ShardBuffer:
    """
    Manages serializing and deserializing shards on the fly.
    Also allows iteration for easy object collection.
    """

    def  __init__(self, table_dir: str, shard_paths: Dict[int, str], config: Config) -> None:
        self.table_dir = table_dir
        self.loaded_shards: Dict[int, Shard] = {}
        self.shard_paths = shard_paths
        self.current_shard_index: int = -1
        self.lru = ShardLRU(max_len=config.page_cache)
        self.config = config

    def __iter__(self):
        self.current_shard_index = -1
        return self

    def __next__(self) -> Shard:
        self.current_shard_index += 1
        if self.current_shard_index in self.shard_paths:
            self._ensure_shard_loaded(self.current_shard_index)
            return self.loaded_shards[self.current_shard_index]
        else:
            raise StopIteration

    def __getitem__(self, shard_index: int) -> Shard:
        self._ensure_shard_loaded(shard_index)
        return self.loaded_shards[shard_index]

    def _create_new_shard_path(self) -> str:
        """Creates a new shard path that will not collide with any others."""
        shard_name = f"shard{len(self.shard_paths)}"
        return os.path.join(self.table_dir, shard_name)

    def _ensure_shard_loaded(self, shard_index: int) -> None:
        """
        Ensures that the given shard index has been loaded.
        :param shard_index:
        :return:
        """
        shard_to_persist = self.lru.update(shard_index)
        if shard_to_persist is not None:
            self._free_shard(shard_to_persist)
        if shard_index not in self.loaded_shards:
            if shard_index in self.shard_paths:
                self.loaded_shards.update(
                    {shard_index: Shard.from_bytes(load_shard(self.shard_paths[shard_index]), self.config.page_size)})
            else:
                self.loaded_shards.update({shard_index: Shard()})
                self.shard_paths.update({shard_index: self._create_new_shard_path()})

    def _free_shard(self, shard: int) -> None:
        """Clears a shard from memory and saves it to disk."""
        self._persist_shard(shard)
        if shard in self.loaded_shards:
            self.loaded_shards.pop(shard)

    def _persist_shard(self, shard: int) -> None:
        """Saves a shard to disk."""
        if not self._shard_has_changes(shard):
            return
        if shard in self.loaded_shards:
            shard_path = self.shard_paths[shard]
            shard_data = self.loaded_shards[shard]
            dump_shard(shard_path, shard_data.to_bytes())

    def _shard_has_changes(self, shard: int) -> bool:
        """Uses checksums to calculate if a shard has changed."""
        if os.path.exists(self.shard_paths[shard]):
            saved_checksum = get_checksum(self.shard_paths[shard])
            return not saved_checksum == self.loaded_shards[shard].checksum
        return True

    def commit(self) -> None:
        """Persists all shards."""
        for shard in self.loaded_shards:
            self._persist_shard(shard)
